import re
from urlparse import urlunsplit, urljoin, uses_relative
from cgi import parse_qsl
from urllib import quote_plus, urlencode
from copy import copy
from mext.dispatch import *
from mext.const import *

__all__ = ['Uri', 'RelativeUri', 'AbsoluteUri', 'urlquote', 'encode_query']

OTHER = const()

def encode_query(*args, **params):
    query = list(args)
    for key, value in params.iteritems():
        value_str = quote_plus(unicode(value).encode('utf-8'))
        query.append("%s=%s" % (key, value_str))
    return '&'.join(query)

def urlquote(*args, **params):
    path = u'/'.join(args).encode('utf8')
    path = quote_plus(path, '/')
    if not path.startswith('/'):
        path = '/' + path
    query = encode_query(**params)
    return (path + '?' + query) if query else path

_parse_cache = {}
_clear_half = True
MAX_CACHE_SIZE = 500


_rx_netloc = re.compile(r'\A([^\/\?\#]+)(.*)\Z')
_rx_http = re.compile(r'\A(http)://([^\/\?\#]+)([^#?]*)(?:\?([^#]*))?(?:#(.*))?\Z')
_rx_relative = re.compile(r'\A([^#?]*)(?:\?([^#]*))?(?:#(.*))?\Z')

scheme_requires_path = set('http https ftp'.split())
scheme_uses_netloc = set((
    'ftp http gopher nntp telnet imap wais file mms '
    'https shttp snews prospero rtsp rtspu rsync '
    'svn svn+ssh sftp').split()
)
scheme_uses_query = set('http https shttp wais imap mms gopher rtsp rtspu sip sips'.split())
scheme_uses_fragment = set('ftp hdl http gopher news nntp wais https shttp snews file prospero'.split())



def _clean_cache():
    global _clear_half
    keys = list(_parse_cache)
    _clear_half = not _clear_half
    idx_half = len(keys)/2
    if _clear_half:
        keys = keys[:idx_half]
    else:
        keys = keys[idx_half:]
    map(_parse_cache.__delitem__, keys)
    del keys


def urlsplit(url):
    """
        Parse a URL into 5 components:
        <scheme>://<netloc></path>?<query>#<fragment>
        Return a 5-tuple: (scheme, netloc, path, query, fragment).
    """
    #url = unicode(url)
    key = type(url), url
    if key in _parse_cache:
        return _parse_cache[key]

    if len(_parse_cache) >= MAX_CACHE_SIZE:
        _clean_cache()

    query = fragment = ''

    if ':' in url:
        if url.startswith('http://'): # optimize the common case
            r = _parse_cache[key] = _rx_http.match(url).groups()
            return r
        scheme, path = url.split(':', 1)
        scheme = scheme.lower()
        if scheme in scheme_uses_netloc:
            if not path.startswith('//'):
                raise ValueError("Invalid URL: %s" % url)
            netloc, path = _rx_netloc.match(path[2:]).groups()
        else:
            netloc = ''
        if scheme in scheme_uses_fragment and '#' in path:
            path, fragment = path.split('#', 1)
        if scheme in scheme_uses_query and '?' in path:
            path, query = path.split('?', 1)
    else:
        scheme = netloc = ''
        path, query, fragment = _rx_relative.match(url).groups()
    r = _parse_cache[key] = (scheme, netloc, path, query, fragment)
    return r




class Uri(object):
    adapt = Adapter('__uri__')

    def __init__(self, *args):
        if len(args) == 1:
            args = urlsplit(unicode(args[0]))
        self.scheme, self.netloc, self.path, self.query, self.fragment = args
        if self._is_absolute:
            cls = AbsoluteUri
            if self.fragment and self.scheme not in scheme_uses_fragment:
                raise ValueError('URI scheme %r does not support fragment identifiers' % self.scheme)
            if not self.path and self.scheme in scheme_requires_path:
                self.path = '/'
        else:
            cls = RelativeUri
        if not isinstance(self, cls):
            if not issubclass(cls, self.__class__):
                raise TypeError("%r is not a subclass of %r" % (cls, self.__class__))
            self.__class__ = cls

    @property
    def parts(self):
        return (self.scheme, self.netloc, self.path, self.query, self.fragment)

    def __str__(self):
        return urlunsplit(self.parts)

    def __repr__(self):
        return self.__class__.__name__ + '(%s)' % repr(str(self))

    @property
    def path_with_query(self):
        path = self.path
        if self.query:
            path += '?' + self.query
        return path

    @classmethod
    def from_parts(cls, scheme=None, netloc=None, path=None, query=None, fragment=None):
        return cls(scheme, netloc, path, query, fragment)

    def copy(self):
        return copy(self)

    def join(self, ref):
        return Uri(urljoin(str(self), str(ref)))

    #@@ maybe just use .join('#frag') ?
    def with_fragment(self, fragment):
        return Uri(self.scheme, self.netloc, self.path, self.query, fragment)

    def as_dir(self, save_query=False):
        r = self.copy()
        if not r.path.endswith('/'):
            r.path += '/'
        if not save_query:
            r.query = None
            r.fragment = None
        return r

    @property
    def _is_absolute(self):
        return bool(self.scheme #and self.path
            and (self.netloc or (self.scheme not in scheme_uses_netloc))
        )

    def defrag(self):
        if self.fragment:
            return self.with_fragment(None)
        else:
            return self


    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return str(self) == other

    def __ne__(self, other):
        return not (self == other)


    def _get_query(self):
        if self._query is OTHER:
            self._query = urlencode(self._query_dict)
            self._query_dict = OTHER
        return self._query
    def _set_query(self, query):
        self._query = query
        self._query_dict = OTHER
    query = property(_get_query, _set_query)


    def _get_query_dict(self):
        if self._query_dict is OTHER:
            self._query_dict = dict(parse_qsl(self._query))
            self._query = OTHER
        return self._query_dict
    def _set_query_dict(self, query_dict):
        self._query = OTHER
        self._query_dict = query_dict
    query_dict = property(_get_query_dict, _set_query_dict)




Uri.adapt[basestring] = Uri
Uri.adapt.register('urllib2.Request', lambda req: AbsoluteUri(req.get_full_url()))

class AbsoluteUri(Uri):
    is_absolute = True
    def relative(self, base):
        base = Uri.adapt(base)
        if not base.is_absolute:
            raise ValueError("Cannon generate relative links from a relative base")
        elif (base.scheme or self.scheme) not in uses_relative:
            return self
        elif self.scheme == base.scheme and self.netloc == base.netloc:
            parts = ['', '', self.path, self.query, self.fragment]
            if base.path == self.path and base.query == self.query:
                parts[2] = parts[3] = ''
            elif '/' in self.path:
                #@@ better relative paths
                #@@ urlparse.non_hierarchical
                base_split = base.path.rsplit('/', 1)
                self_split = self.path.rsplit('/', 1)
                if base_split[0] == self_split[0]:
                    parts[2] = self_split[1]
            return RelativeUri(*parts)
            #assert base.join(ref) == self, repr((base, ref, base.join(ref), self))
        else:
            return self


class RelativeUri(Uri):
    is_absolute = False

    def __str__(self):
        if (self.scheme or self.netloc or self.path or self.query):
            return Uri.__str__(self)
        else:
            return '#%s' % (self.fragment or '')

    def __nonzero__(self):
        return bool(self.scheme or self.netloc or self.path or self.query or self.fragment)






